//! Quick and dirty uart driver

use core::fmt::Write;
/// TODO repr(C) a register bank, and add instances
use core::ptr::{read_volatile, write_volatile};

const REG_RXTX: u32 = 0x00;
const REG_TXFULL: u32 = 0x04;
const REG_RXEMPTY: u32 = 0x08;
const REG_TXEMPTY: u32 = 0x18;
const REG_RXFULL: u32 = 0x1c;

pub enum Error {
    TxFull,
    RxEmpty,
}

// Hacky derive, shouldn't exist but it's fine :tm:
#[derive(Copy, Clone)]
pub struct LitexUart {
    base_addr: u32,
}

impl LitexUart {
    pub fn new(base_addr: u32) -> Self {
        Self { base_addr }
    }

    unsafe fn read_reg(&mut self, offset: u32) -> u32 {
        read_volatile((self.base_addr + offset) as *const u32)
    }

    pub fn try_get_char(&mut self) -> Result<u8, Error> {
        unsafe {
            if self.read_reg(REG_RXEMPTY) != 0 {
                return Err(Error::RxEmpty);
            }

            Ok(read_volatile((self.base_addr + REG_RXTX) as *mut u8))
        }
    }

    pub fn try_put_char(&mut self, c: u8) -> Result<(), Error> {
        unsafe {
            if self.read_reg(REG_TXFULL) != 0 {
                return Err(Error::TxFull);
            }

            write_volatile((self.base_addr + REG_RXTX) as *mut u8, c);
            Ok(())
        }
    }
}

// Blocking implementation of write
impl Write for LitexUart {
    fn write_str(&mut self, s: &str) -> core::fmt::Result {
        for b in s.as_bytes() {
            // It's okay to loop on this because we'll always clear the buffer
            while let Err(Error::TxFull) = self.try_put_char(*b) {}
            //self.try_put_char(*b);
        }

        Ok(())
    }
}
